/*

  Broadcom B43 wireless driver

  Transmission (TX/RX) related functions.

  Copyright (C) 2005 Martin Langer <martin-langer@gmx.de>
  Copyright (C) 2005 Stefano Brivio <stefano.brivio@polimi.it>
  Copyright (C) 2005, 2006 Michael Buesch <mb@bu3sch.de>
  Copyright (C) 2005 Danny van Dyk <kugelfang@gentoo.org>
  Copyright (C) 2005 Andreas Jaggi <andreas.jaggi@waterwave.ch>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, Inc., 51 Franklin Steet, Fifth Floor,
  Boston, MA 02110-1301, USA.

*/

#include <linux/crc32.h>
#include "xmit.h"
#include "phy_common.h"
#include "dma.h"
#include "pio.h"

int rateTot = 2; // for debugging purposes

/* GRINGO define MULTI11_DEBUG to printk several debug messages */
#define noMULTI11_DEBUG

/* GRINGO define FLETCHER32 to use fletcher32 instead of crc32_be
   to compute the crc coefficient for final block aggregation
*/
#define FLETCHER32
#define FLETCHER_BLOCK_SIZE (blockLength)

/* GRINGO we use this function to increase the rate for testing purposes */
int b43_test_double_rate(int rate)
{
	switch(rate) {
	case 2:	// was 1
		return 4;
	case 4: // was 2
		return 11;
	case 11: // was 5.5
		return 22;
	case 22: // was 11
		return 12;
	case 12: // was 6
		return 18;
	case 18: // was 9
		return 24;
	case 24: // was 12
		return 36;
	case 36: // was 18
		return 48;
	case 48: // was 24
		return 72;
	case 72: // was 36
		return 96;
	case 96: // was 48
		return 108;
	default: 
		return 2;
	}
	return 2;
}

/* GRINGO we put this helper to index the various rates at 2.4GHz
   Note: it is a qualitative index */
static u8 b43_get_index_from_rate(const u8 bitrate)
{
	switch(bitrate) {
       case B43_CCK_RATE_1MB:
               return 0;
       case B43_CCK_RATE_2MB:
                return 1;
        case B43_CCK_RATE_5MB:
                return 2;
        case B43_CCK_RATE_11MB:
                return 3;
        case B43_OFDM_RATE_6MB:
		return 4;
       case B43_OFDM_RATE_9MB:
	return 5;
      case B43_OFDM_RATE_12MB:
		return 6;
        case B43_OFDM_RATE_18MB:
		return 7;
       case B43_OFDM_RATE_24MB:
	return 8;
        case B43_OFDM_RATE_36MB:
		return 9;
        case B43_OFDM_RATE_48MB:
		return 10;
        case B43_OFDM_RATE_54MB:
		return 11;
	}
        B43_WARN_ON(1);
        return 0;
}

/* GRINGO: another helper, provides a translation between hardware ratecode
           and channel speed */
static u8 b43_get_speed_from_cck_ratecode(const u8 ratecode)
{
	switch(ratecode) {
	case 0x0a:
		return B43_CCK_RATE_1MB;
	case 0x14:
		return B43_CCK_RATE_2MB;
	case 0x37:
		return B43_CCK_RATE_5MB;
	case 0x6e:
		return B43_CCK_RATE_11MB;
	}
	B43_WARN_ON(1);
	return 0;
}

static u8 b43_get_speed_from_ofdm_ratecode(const u8 ratecode)
{
	switch(ratecode) {
	case 0x0b:
		return B43_OFDM_RATE_6MB;
	case 0x0f:
		return B43_OFDM_RATE_9MB;
	case 0x0a:
		return B43_OFDM_RATE_12MB;
	case 0x0e:
		return B43_OFDM_RATE_18MB;
	case 0x09:
		return B43_OFDM_RATE_24MB;
	case 0x0d:
		return B43_OFDM_RATE_36MB;
	case 0x08:
		return B43_OFDM_RATE_48MB;
	case 0x0c:
		return B43_OFDM_RATE_54MB;
	}
	B43_WARN_ON(1);
	return 0;
}


/* Extract the bitrate index out of a CCK PLCP header. */
static int b43_plcp_get_bitrate_idx_cck(struct b43_plcp_hdr6 *plcp)
{
	switch (plcp->raw[0]) {
	case 0x0A:
		return 0;
	case 0x14:
		return 1;
	case 0x37:
		return 2;
	case 0x6E:
		return 3;
	}
	return -1;
}

/* Extract the bitrate index out of an OFDM PLCP header. */
static int b43_plcp_get_bitrate_idx_ofdm(struct b43_plcp_hdr6 *plcp, bool aphy)
{
	int base = aphy ? 0 : 4;

	switch (plcp->raw[0] & 0xF) {
	case 0xB:
		return base + 0;
	case 0xF:
		return base + 1;
	case 0xA:
		return base + 2;
	case 0xE:
		return base + 3;
	case 0x9:
		return base + 4;
	case 0xD:
		return base + 5;
	case 0x8:
		return base + 6;
	case 0xC:
		return base + 7;
	}
	return -1;
}

u8 b43_plcp_get_ratecode_cck(const u8 bitrate)
{
	switch (bitrate) {
	case B43_CCK_RATE_1MB:
		return 0x0A;
	case B43_CCK_RATE_2MB:
		return 0x14;
	case B43_CCK_RATE_5MB:
		return 0x37;
	case B43_CCK_RATE_11MB:
		return 0x6E;
	}
	B43_WARN_ON(1);
	return 0;
}

u8 b43_plcp_get_ratecode_ofdm(const u8 bitrate)
{
	switch (bitrate) {
	case B43_OFDM_RATE_6MB:
		return 0xB;
	case B43_OFDM_RATE_9MB:
		return 0xF;
	case B43_OFDM_RATE_12MB:
		return 0xA;
	case B43_OFDM_RATE_18MB:
		return 0xE;
	case B43_OFDM_RATE_24MB:
		return 0x9;
	case B43_OFDM_RATE_36MB:
		return 0xD;
	case B43_OFDM_RATE_48MB:
		return 0x8;
	case B43_OFDM_RATE_54MB:
		return 0xC;
	}
	B43_WARN_ON(1);
	return 0;
}

void b43_generate_plcp_hdr(struct b43_plcp_hdr4 *plcp,
			   const u16 octets, const u8 bitrate)
{
	__u8 *raw = plcp->raw;

	if (b43_is_ofdm_rate(bitrate)) {
		u32 d;

		d = b43_plcp_get_ratecode_ofdm(bitrate);
		B43_WARN_ON(octets & 0xF000);
		d |= (octets << 5);
		plcp->data = cpu_to_le32(d);
	} else {
		u32 plen;

		plen = octets * 16 / bitrate;
		if ((octets * 16 % bitrate) > 0) {
			plen++;
			if ((bitrate == B43_CCK_RATE_11MB)
			    && ((octets * 8 % 11) < 4)) {
				raw[1] = 0x84;
			} else
				raw[1] = 0x04;
		} else
			raw[1] = 0x04;
		plcp->data |= cpu_to_le32(plen << 16);
		raw[0] = b43_plcp_get_ratecode_cck(bitrate);
	}
}

static u8 b43_calc_fallback_rate(u8 bitrate)
{
	switch (bitrate) {
	case B43_CCK_RATE_1MB:
		return B43_CCK_RATE_1MB;
	case B43_CCK_RATE_2MB:
		return B43_CCK_RATE_1MB;
	case B43_CCK_RATE_5MB:
		return B43_CCK_RATE_2MB;
	case B43_CCK_RATE_11MB:
		return B43_CCK_RATE_5MB;
	case B43_OFDM_RATE_6MB:
		return B43_CCK_RATE_5MB;
	case B43_OFDM_RATE_9MB:
		return B43_OFDM_RATE_6MB;
	case B43_OFDM_RATE_12MB:
		return B43_OFDM_RATE_9MB;
	case B43_OFDM_RATE_18MB:
		return B43_OFDM_RATE_12MB;
	case B43_OFDM_RATE_24MB:
		return B43_OFDM_RATE_18MB;
	case B43_OFDM_RATE_36MB:
		return B43_OFDM_RATE_24MB;
	case B43_OFDM_RATE_48MB:
		return B43_OFDM_RATE_36MB;
	case B43_OFDM_RATE_54MB:
		return B43_OFDM_RATE_48MB;
	}
	B43_WARN_ON(1);
	return 0;
}

// please note: we do not use standard 720-bytes block sizes because we need
// to have a block size for the global coefficient that is a multiple of
// the basic block size (704 and 64 respectively)
__le32 fletcher32(const unsigned char *block_data,
			int fragment_len,
			u8 *crc_bank)
{
	u32 sum1 = 0xffff, sum2 = 0xffff;
	__le32 coeff64;
	unsigned char* new_block_data = (unsigned char*) block_data;
	int bank_done = 0;

	// determine the maximum number of bytes before overflow
	// normally this is 720 so we must choose a number
	// that is multiple of the basic block size
	// FLETCHER_BLOCK_SIZE and smaller than 720
	int magic_number = 720 - (720 % FLETCHER_BLOCK_SIZE);

	// skip only if we need to, e.g., if the caller wants all the coefficients
	// skip the first coefficient, it will be up to the caller to store there
	// the global coefficient returned by this function.
	if(crc_bank != NULL)
		crc_bank += 4;

	// fragment_len should be smaller than 1538
	// avoid computation of coefficients in these cases as the firmware will not undergo special retransmissions
	if(fragment_len <= 58 || fragment_len > 1538) {
		return 0;
	}

	// skip the first 58 bytes
	fragment_len -= 58;
	new_block_data += 58;

	// loop until we analyzed all the bytes
	while (fragment_len > 0) {
		unsigned tlen = fragment_len > magic_number ? magic_number : fragment_len;
		fragment_len -= tlen;

		// inside this code block we have to compute the coefficient
		// for tlen bytes
		// produce a new coefficient for every FLETCHER_BLOCK_SIZE bytes and
		// accumulate over tlen bytes for the global one
		while (tlen > 0) {
			unsigned nlen = tlen > FLETCHER_BLOCK_SIZE ? FLETCHER_BLOCK_SIZE : tlen;
			u32 sum3 = 0xffff;
			u32 sum4 = 0xffff;

			tlen -= nlen;

			while(nlen > 1) {
				sum1 += (new_block_data[0] & 0xff) | ((new_block_data[1] << 8) & 0xff00);
				sum2 += sum1;
				sum3 += (new_block_data[0] & 0xff) | ((new_block_data[1] << 8) & 0xff00);
				sum4 += sum3;

				new_block_data += 2;
				nlen -= 2;
			}

			// this case happens only once, do not refresh data
			if(nlen == 1) {
				sum1 += (new_block_data[0] & 0xff);
				sum2 += sum1;
				sum3 += (new_block_data[0] & 0xff);
				sum4 += sum3;
			}

			// compute the coefficient for this 64-bytes block
			sum3 = (sum3 & 0xffff) + (sum3 >> 16);
			sum4 = (sum4 & 0xffff) + (sum4 >> 16);
			coeff64 = sum4 << 16 | sum3;

			// check we are within the correct number of bank
			if(bank_done > 24) {
				B43_WARN_ON(1);
				//printk("too many bank already done\n");
				return 0;
			}
			bank_done++;

			if(crc_bank != NULL) {
				crc_bank[0] = coeff64 & 0xff;
				crc_bank[1] = (coeff64 >> 8) & 0xff;
				crc_bank[2] = (coeff64 >> 16) & 0xff;
				crc_bank[3] = (coeff64 >> 24) & 0xff;
				crc_bank += 4;
			}
		}

		// wrap the global coefficient
		sum1 = (sum1 & 0xffff) + (sum1 >> 16);
		sum2 = (sum2 & 0xffff) + (sum2 >> 16);
	}

	// finalize the global coefficient
	sum1 = (sum1 & 0xffff) + (sum1 >> 16);
	sum2 = (sum2 & 0xffff) + (sum2 >> 16);
	coeff64 = sum2 << 16 | sum1;

	return coeff64;
}


/* Generate a TX data header. */
int b43_generate_txhdr(struct b43_wldev *dev,
		       u8 *_txhdr,
		       const unsigned char *fragment_data,
		       unsigned int fragment_len,
		       struct ieee80211_tx_info *info,
		       u16 cookie)
{
	struct b43_txhdr *txhdr = (struct b43_txhdr *)_txhdr;
	const struct b43_phy *phy = &dev->phy;
	const struct ieee80211_hdr *wlhdr =
	    (const struct ieee80211_hdr *)fragment_data;
	int use_encryption = !!info->control.hw_key;
	__le16 fctl = wlhdr->frame_control;
	struct ieee80211_rate *fbrate;
	u8 rate, rate_fb;
	int rate_ofdm, rate_fb_ofdm;
	unsigned int plcp_fragment_len;
	u32 mac_ctl = 0;
	u16 phy_ctl = 0;
	u8 extra_ft = 0;
	struct ieee80211_rate *txrate;
	struct ieee80211_tx_rate *rates;

	u32 extra_crc32;
	int fakeNoAck = 0;
	int rate_idx;

	__le16* crc_bank;
	u8* crc32;

	memset(txhdr, 0, sizeof(*txhdr));

	txrate = ieee80211_get_tx_rate(dev->wl->hw, info);
	rate = txrate ? txrate->hw_value : B43_CCK_RATE_1MB;

        if (b43_is_old_txhdr_format(dev))
                crc_bank = &txhdr->old_format.crc_bank[0];
        else
                crc_bank = &txhdr->new_format.crc_bank[0];

	crc32=(u8*)crc_bank;

	// compute all the coefficients
	extra_crc32 = fletcher32(fragment_data, fragment_len, (u8*)crc_bank);
#ifndef FLETCHER32
	// overwrite the previous value. It's computation has not slow
	// down since it is automatically computed given that we must
	// compute partial coefficients
	extra_crc32 = crc32_be( ~0, fragment_data + 32, fragment_len - 32);
#endif

	/* check if power flag is set then send with special retransmission beginning with
		rate chosen during module load */
	if(fragment_len == 1530) {
		struct ieee80211_hdr *wlhdr_bis = (struct ieee80211_hdr *) wlhdr;
		wlhdr_bis->frame_control &= ~0x1000;
		// set first two bytes of first crc to carry the crc32 of the fake packet
		// fragment_len does not include the tailing crc, hence we should compute
		// the crc for everything following the llc layer
		crc32[0] = extra_crc32 & 0xFF;
		crc32[1] = (extra_crc32 >> 8) & 0xFF;
		crc32[2] = (extra_crc32 >> 16) & 0xFF;
		crc32[3] = (extra_crc32 >> 24) & 0xFF;
		rate = rateFakePacket;
		/* uncomment this to avoid waiting for an ack (at firmware) */
		//fakeNoAck = 1;
	}

	/* if this packet should be transmitted using maranello then enable the firmware */
	if  (fragment_len == 1530) {
		crc32[0] = extra_crc32 & 0xFF;
		crc32[1] = (extra_crc32 >> 8) & 0xFF;
		crc32[2] = (extra_crc32 >> 16) & 0xFF;
		crc32[3] = (extra_crc32 >> 24) & 0xFF;
		//rate = rateTot;
		//rateTot = b43_test_double_rate(rateTot);

	}

	//Added by Bo Han for rate control, 2009-03-24
  txrate = ieee80211_get_tx_rate(dev->wl->hw, info);
  rate = txrate ? txrate->hw_value : B43_CCK_RATE_1MB;

	
	// refresh the number of frames sent for legacy or maranello mac
	if(crc32[0] != 0) {
		dev->maranello.sent++;
	} else {
		dev->legacy.sent++;
 	}

	// refresh the number of frames sent with a given first rate
	rate_idx = b43_get_index_from_rate(rate);
	if(crc32[0] != 0) {
		dev->maranello.sent_by_rate[rate_idx]++;
	} else {
		dev->legacy.sent_by_rate[rate_idx]++;
	}

	rate_ofdm = b43_is_ofdm_rate(rate);
	fbrate = ieee80211_get_alt_retry_rate(dev->wl->hw, info, 0) ? : txrate;
	rate_fb = fbrate->hw_value;
	rate_fb_ofdm = b43_is_ofdm_rate(rate_fb);

	if (rate_ofdm)
		txhdr->phy_rate = b43_plcp_get_ratecode_ofdm(rate);
	else
		txhdr->phy_rate = b43_plcp_get_ratecode_cck(rate);
	txhdr->mac_frame_ctl = wlhdr->frame_control;
	memcpy(txhdr->tx_receiver, wlhdr->addr1, 6);

	/* Calculate duration for fallback rate */
	if ((rate_fb == rate) ||
	    (wlhdr->duration_id & cpu_to_le16(0x8000)) ||
	    (wlhdr->duration_id == cpu_to_le16(0))) {
		/* If the fallback rate equals the normal rate or the
		 * dur_id field contains an AID, CFP magic or 0,
		 * use the original dur_id field. */
		txhdr->dur_fb = wlhdr->duration_id;
	} else {
		txhdr->dur_fb = ieee80211_generic_frame_duration(
			dev->wl->hw, info->control.vif, fragment_len, fbrate);
	}

	plcp_fragment_len = fragment_len + FCS_LEN;
	if (use_encryption) {
		u8 key_idx = info->control.hw_key->hw_key_idx;
		struct b43_key *key;
		int wlhdr_len;
		size_t iv_len;

		B43_WARN_ON(key_idx >= dev->max_nr_keys);
		key = &(dev->key[key_idx]);

		if (unlikely(!key->keyconf)) {
			/* This key is invalid. This might only happen
			 * in a short timeframe after machine resume before
			 * we were able to reconfigure keys.
			 * Drop this packet completely. Do not transmit it
			 * unencrypted to avoid leaking information. */
			return -ENOKEY;
		}

		/* Hardware appends ICV. */
		plcp_fragment_len += info->control.hw_key->icv_len;

		key_idx = b43_kidx_to_fw(dev, key_idx);
		mac_ctl |= (key_idx << B43_TXH_MAC_KEYIDX_SHIFT) &
			   B43_TXH_MAC_KEYIDX;
		mac_ctl |= (key->algorithm << B43_TXH_MAC_KEYALG_SHIFT) &
			   B43_TXH_MAC_KEYALG;
		wlhdr_len = ieee80211_hdrlen(fctl);
		iv_len = min((size_t) info->control.hw_key->iv_len,
			     ARRAY_SIZE(txhdr->iv));
		memcpy(txhdr->iv, ((u8 *) wlhdr) + wlhdr_len, iv_len);
	}
	if (b43_is_old_txhdr_format(dev)) {
		b43_generate_plcp_hdr((struct b43_plcp_hdr4 *)(&txhdr->old_format.plcp),
				      plcp_fragment_len, rate);
	} else {
		b43_generate_plcp_hdr((struct b43_plcp_hdr4 *)(&txhdr->new_format.plcp),
				      plcp_fragment_len, rate);
		/* we use this field to tell the firmware about the actual length of the frame */
		txhdr->new_format.frame_length = plcp_fragment_len;	
	}
	b43_generate_plcp_hdr((struct b43_plcp_hdr4 *)(&txhdr->plcp_fb),
			      plcp_fragment_len, rate_fb);

	/* Extra Frame Types */
	if (rate_fb_ofdm)
		extra_ft |= B43_TXH_EFT_FB_OFDM;
	else
		extra_ft |= B43_TXH_EFT_FB_CCK;

	/* Set channel radio code. Note that the micrcode ORs 0x100 to
	 * this value before comparing it to the value in SHM, if this
	 * is a 5Ghz packet.
	 */
	txhdr->chan_radio_code = phy->channel;

	/* PHY TX Control word */
	if (rate_ofdm)
		phy_ctl |= B43_TXH_PHY_ENC_OFDM;
	else
		phy_ctl |= B43_TXH_PHY_ENC_CCK;
	if (info->control.rates[0].flags & IEEE80211_TX_RC_USE_SHORT_PREAMBLE)
		phy_ctl |= B43_TXH_PHY_SHORTPRMBL;

	switch (b43_ieee80211_antenna_sanitize(dev, info->antenna_sel_tx)) {
	case 0: /* Default */
		phy_ctl |= B43_TXH_PHY_ANT01AUTO;
		break;
	case 1: /* Antenna 0 */
		phy_ctl |= B43_TXH_PHY_ANT0;
		break;
	case 2: /* Antenna 1 */
		phy_ctl |= B43_TXH_PHY_ANT1;
		break;
	case 3: /* Antenna 2 */
		phy_ctl |= B43_TXH_PHY_ANT2;
		break;
	case 4: /* Antenna 3 */
		phy_ctl |= B43_TXH_PHY_ANT3;
		break;
	default:
		B43_WARN_ON(1);
	}

	rates = info->control.rates;
	/* MAC control */
	if (!(info->flags & IEEE80211_TX_CTL_NO_ACK))
		mac_ctl |= B43_TXH_MAC_ACK;

	/* do not wait ACK */
	if(fakeNoAck)
		mac_ctl &= ~B43_TXH_MAC_ACK;

	/* use hardware sequence counter as the non-TID counter */
	if (info->flags & IEEE80211_TX_CTL_ASSIGN_SEQ)
		mac_ctl |= B43_TXH_MAC_HWSEQ;
	if (info->flags & IEEE80211_TX_CTL_FIRST_FRAGMENT)
		mac_ctl |= B43_TXH_MAC_STMSDU;
	if (phy->type == B43_PHYTYPE_A)
		mac_ctl |= B43_TXH_MAC_5GHZ;

	/* Overwrite rates[0].count to make the retry calculation
	 * in the tx status easier. need the actual retry limit to
	 * detect whether the fallback rate was used.
	 */
	if ((rates[0].flags & IEEE80211_TX_RC_USE_RTS_CTS) ||
	    (rates[0].count <= dev->wl->hw->conf.long_frame_max_tx_count)) {
		rates[0].count = dev->wl->hw->conf.long_frame_max_tx_count;
		mac_ctl |= B43_TXH_MAC_LONGFRAME;
	} else {
		rates[0].count = dev->wl->hw->conf.short_frame_max_tx_count;
	}

	/* Generate the RTS or CTS-to-self frame */
	if ((rates[0].flags & IEEE80211_TX_RC_USE_RTS_CTS) ||
	    (rates[0].flags & IEEE80211_TX_RC_USE_CTS_PROTECT)) {
		unsigned int len;
		struct ieee80211_hdr *hdr;
		int rts_rate, rts_rate_fb;
		int rts_rate_ofdm, rts_rate_fb_ofdm;
		struct b43_plcp_hdr6 *plcp;
		struct ieee80211_rate *rts_cts_rate;

		rts_cts_rate = ieee80211_get_rts_cts_rate(dev->wl->hw, info);

		rts_rate = rts_cts_rate ? rts_cts_rate->hw_value : B43_CCK_RATE_1MB;
		rts_rate_ofdm = b43_is_ofdm_rate(rts_rate);
		rts_rate_fb = b43_calc_fallback_rate(rts_rate);
		rts_rate_fb_ofdm = b43_is_ofdm_rate(rts_rate_fb);

		if (rates[0].flags & IEEE80211_TX_RC_USE_CTS_PROTECT) {
			struct ieee80211_cts *cts;

			if (b43_is_old_txhdr_format(dev)) {
				cts = (struct ieee80211_cts *)
					(txhdr->old_format.rts_frame);
			} else {
				cts = (struct ieee80211_cts *)
					(txhdr->new_format.rts_frame);
			}
			ieee80211_ctstoself_get(dev->wl->hw, info->control.vif,
						fragment_data, fragment_len,
						info, cts);
			mac_ctl |= B43_TXH_MAC_SENDCTS;
			len = sizeof(struct ieee80211_cts);
		} else {
			struct ieee80211_rts *rts;

			if (b43_is_old_txhdr_format(dev)) {
				rts = (struct ieee80211_rts *)
					(txhdr->old_format.rts_frame);
			} else {
				rts = (struct ieee80211_rts *)
					(txhdr->new_format.rts_frame);
			}
			ieee80211_rts_get(dev->wl->hw, info->control.vif,
					  fragment_data, fragment_len,
					  info, rts);
			mac_ctl |= B43_TXH_MAC_SENDRTS;
			len = sizeof(struct ieee80211_rts);
		}
		len += FCS_LEN;

		/* Generate the PLCP headers for the RTS/CTS frame */
		if (b43_is_old_txhdr_format(dev))
			plcp = &txhdr->old_format.rts_plcp;
		else
			plcp = &txhdr->new_format.rts_plcp;
		b43_generate_plcp_hdr((struct b43_plcp_hdr4 *)plcp,
				      len, rts_rate);
		plcp = &txhdr->rts_plcp_fb;
		b43_generate_plcp_hdr((struct b43_plcp_hdr4 *)plcp,
				      len, rts_rate_fb);

		if (b43_is_old_txhdr_format(dev)) {
			hdr = (struct ieee80211_hdr *)
				(&txhdr->old_format.rts_frame);
		} else {
			hdr = (struct ieee80211_hdr *)
				(&txhdr->new_format.rts_frame);
		}
		txhdr->rts_dur_fb = hdr->duration_id;

		if (rts_rate_ofdm) {
			extra_ft |= B43_TXH_EFT_RTS_OFDM;
			txhdr->phy_rate_rts =
			    b43_plcp_get_ratecode_ofdm(rts_rate);
		} else {
			extra_ft |= B43_TXH_EFT_RTS_CCK;
			txhdr->phy_rate_rts =
			    b43_plcp_get_ratecode_cck(rts_rate);
		}
		if (rts_rate_fb_ofdm)
			extra_ft |= B43_TXH_EFT_RTSFB_OFDM;
		else
			extra_ft |= B43_TXH_EFT_RTSFB_CCK;
	}

	/* Magic cookie */
	if (b43_is_old_txhdr_format(dev))
		txhdr->old_format.cookie = cpu_to_le16(cookie);
	else
		txhdr->new_format.cookie = cpu_to_le16(cookie);

	/* Apply the bitfields */
	txhdr->mac_ctl = cpu_to_le32(mac_ctl);
	txhdr->phy_ctl = cpu_to_le16(phy_ctl);
	txhdr->extra_ft = extra_ft;

	return 0;
}

static s8 b43_rssi_postprocess(struct b43_wldev *dev,
			       u8 in_rssi, int ofdm,
			       int adjust_2053, int adjust_2050)
{
	struct b43_phy *phy = &dev->phy;
	struct b43_phy_g *gphy = phy->g;
	s32 tmp;

	switch (phy->radio_ver) {
	case 0x2050:
		if (ofdm) {
			tmp = in_rssi;
			if (tmp > 127)
				tmp -= 256;
			tmp *= 73;
			tmp /= 64;
			if (adjust_2050)
				tmp += 25;
			else
				tmp -= 3;
		} else {
			if (dev->dev->bus->sprom.
			    boardflags_lo & B43_BFL_RSSI) {
				if (in_rssi > 63)
					in_rssi = 63;
				B43_WARN_ON(phy->type != B43_PHYTYPE_G);
				tmp = gphy->nrssi_lt[in_rssi];
				tmp = 31 - tmp;
				tmp *= -131;
				tmp /= 128;
				tmp -= 57;
			} else {
				tmp = in_rssi;
				tmp = 31 - tmp;
				tmp *= -149;
				tmp /= 128;
				tmp -= 68;
			}
			if (phy->type == B43_PHYTYPE_G && adjust_2050)
				tmp += 25;
		}
		break;
	case 0x2060:
		if (in_rssi > 127)
			tmp = in_rssi - 256;
		else
			tmp = in_rssi;
		break;
	default:
		tmp = in_rssi;
		tmp -= 11;
		tmp *= 103;
		tmp /= 64;
		if (adjust_2053)
			tmp -= 109;
		else
			tmp -= 83;
	}

	return (s8) tmp;
}

//TODO
#if 0
static s8 b43_rssinoise_postprocess(struct b43_wldev *dev, u8 in_rssi)
{
	struct b43_phy *phy = &dev->phy;
	s8 ret;

	if (phy->type == B43_PHYTYPE_A) {
		//TODO: Incomplete specs.
		ret = 0;
	} else
		ret = b43_rssi_postprocess(dev, in_rssi, 0, 1, 1);

	return ret;
}
#endif

u8 get_speed_from_encoding(u8 b0, u8 b1)
{
	/* b0 and b1 are the first and second byte from the plcp
		if b0<<8 | b1 gives one of the following in the first column, then
		assign the speed as in the second column (multiplied by 2)
	0x040A 1
	0x0414 2
	0x0437 5.5
	0x046e 11
	0x846e 11

	otherwise check to see if b0 & 0xf match one of the following
	[6, 9, 12, 18, 24, 36, 48, 54] => [0xB, 0xF, 0xA, 0xE, 0x9, 0xD, 0x8, 0xC]

	*/

	u8 speed;
	u16 encoding = (b1 << 8) | b0;
	switch(encoding) {
	case 0x040A:
		speed = 2;
		break;
	case 0x0414:
		speed = 4;
		break;
	case 0x0437:
		speed = 11;
		break;
	case 0x046E:
	case 0x846E:
		speed = 22;
		break;
	default:
		switch(b0 & 0xF) {
		case 0xB:
			speed = 12;
			break;
		case 0xF:
			speed = 18;
			break;
		case 0xA:
			speed = 24;
			break;
		case 0xE:
			speed = 36;
			break;
		case 0x9:
			speed = 48;
			break;
		case 0xD:
			speed = 72;
			break;
		case 0x8:
			speed = 96;
			break;
		case 0xC:
			speed = 108;
			break;
		default:
			speed = 0;
		}
	}

	return speed;
}


void b43_rx(struct b43_wldev *dev, struct sk_buff *skb, const void *_rxhdr)
{
	struct ieee80211_rx_status status;
	struct b43_plcp_hdr6 *plcp;
	struct ieee80211_hdr *wlhdr;
	const struct b43_rxhdr_fw4 *rxhdr = _rxhdr;
	__le16 fctl;
	u16 phystat0, phystat3, chanstat, mactime;
	u32 macstat;
	u16 chanid;
	u16 phytype;
	int padding;
	u16 fakebad; 
	u8* llc;
	int ii, ones;
	long int m, sigma_tmp, sum_sigma;
	int tmp_mask_vector[25];

	memset(&status, 0, sizeof(status));

	/* Get metadata about the frame from the header. */
	phystat0 = le16_to_cpu(rxhdr->phy_status0);
	phystat3 = le16_to_cpu(rxhdr->phy_status3);
	macstat = le32_to_cpu(rxhdr->mac_status);
	mactime = le16_to_cpu(rxhdr->mac_time);
	chanstat = le16_to_cpu(rxhdr->channel);
	fakebad = le16_to_cpu(rxhdr->fakebad);
	phytype = chanstat & B43_RX_CHAN_PHYTYPE;

	// fakebad == 0x02 should be evaluated first: it is a special code
	// generated by the firmware to tell us that the received
	// frame is due to a correctly received ack (and it is the ack)
	// but all the rxheaders have been replaced to incorporate
	// other kind of information (read the code here below)
	if (fakebad == 0x02) {
		int delay, sent_packet_length;
		int ack_ratecode, pkt_ratecode;
		int ack_speed, pkt_speed;
		int was_maranello = 0;
		delay = ( (chanstat << 16) & 0xffff0000) | mactime;
		ack_ratecode = (macstat >> 16) & 0xffff;
		sent_packet_length = (macstat & 0x7fff);
		if(macstat & 0x8000) {
			was_maranello = 1;
		}
		pkt_ratecode = phystat3;

		// get ack rate speed
		if(ack_ratecode & 0x0100) {
			// OFDM
			ack_speed = b43_get_speed_from_ofdm_ratecode(ack_ratecode & 0x0f);
		} else {
			// CCK
			ack_speed = b43_get_speed_from_cck_ratecode(ack_ratecode & 0xff);
		}

		// get pkt rate speed
		if(pkt_ratecode & 0x0100) {
			// OFDM
			pkt_speed = b43_get_speed_from_ofdm_ratecode(pkt_ratecode & 0x0f);
		} else {
			// CCK
			pkt_speed = b43_get_speed_from_cck_ratecode(pkt_ratecode & 0xff);
		}

		//printk("pkt tx time = %08X, ack_speed = %08X, pkt_speed = %08X, sent_length = %08X, was ZTMD = %d\n", delay, ack_speed, pkt_speed, sent_packet_length, was_maranello);
		if(was_maranello) {
			dev->maranello.stat++;
			dev->maranello.time += delay;
		} else {
			dev->legacy.stat++;
			dev->legacy.time += delay;
		}
		
		dev_kfree_skb_any(skb);
		dev->rtt_measurements ++;
		return;
	}

	if (macstat & B43_RX_MAC_FCSERR) {
		b43dbg(dev->wl, "FCS error\n");
		dev->wl->ieee_stats.dot11FCSErrorCount++;
		status.flag |= RX_FLAG_FAILED_FCS_CRC;
	}
	// multi11: I'm not sure if we should keep the frame
	if (unlikely(phystat0 & (B43_RX_PHYST0_PLCPHCF | B43_RX_PHYST0_PLCPFV)))
		status.flag |= RX_FLAG_FAILED_PLCP_CRC;
	if (phystat0 & B43_RX_PHYST0_SHORTPRMBL)
		status.flag |= RX_FLAG_SHORTPRE;
	if (macstat & B43_RX_MAC_DECERR) {
		/* Decryption with the given key failed.
		 * Drop the packet. We also won't be able to decrypt it with
		 * the key in software. */
		goto drop;
	}

	/* Skip PLCP and padding */
	padding = (macstat & B43_RX_MAC_PADDING) ? 2 : 0;
	if (unlikely(skb->len < (sizeof(struct b43_plcp_hdr6) + padding))) {
		b43dbg(dev->wl, "RX: Packet size underrun (1)\n");
		goto drop;
	}
	// multi11: store the skb and the padding
	// we should first determine in which position we are storing the frame
	if ((fakebad == 0x01) && !dev->wl->radiotap_enabled) {
                int rp;
                unsigned long mintime = jiffies;
                int minpos = 0;
                int nullpos = MAX_SPEC_BLOCKS;

		// check that the frame contains at least the source address of the sender station
		// if shorter do not consider this frame for special retransmission
		// (firmware will not send nack for short frames)
		if(unlikely(skb->len - sizeof(struct b43_plcp_hdr6) - padding <
			(2 + 2 + 18 /*minimum hdr with three addresses */  + FCS_LEN) )) {
#ifdef MULTI11_DEBUG
			b43dbg(dev->wl, "block with FCS error is too short, discarding...\n");
#endif
			goto drop;
		}

                // if we are on an AP we should handle multiple stations,
                // otherwise just keep one skb
                // Note: if we have a fakebad frame from a station (AP case)
                // for which we already hold a skb inside the skb array, then
                // we should remove the old frame and replace it with the new (this one)
                if(b43_is_mode(dev->wl, NL80211_IFTYPE_AP)) {

                        // analyze this frame, we already know we have enough data
                        struct ieee80211_hdr *this_wlhdr = (struct ieee80211_hdr *)(skb->data + padding +
                                sizeof(struct b43_plcp_hdr6));

                        // loop to determine
                        // - if we already have a skb from this sender (replace with the new)
                        // - the oldest skb if the array is full
                        // - a free pointer inside the array
                        for(rp = 0; rp < MAX_SPEC_BLOCKS; rp++) {
                                struct ieee80211_hdr* check_wlhdr;
                                int check_padding;
                                struct sk_buff* check_skb;

                                // no frame here, keep this position and loop again
                                if(dev->skb_fake_bad[rp].skb == NULL) {
                                        if(nullpos == MAX_SPEC_BLOCKS)
                                                nullpos = rp;
                                        continue;
                                }

                                // analyze the frame at this position inside the array
				// we are sure it contains the three addresses as only
				// such frames are kept (see above)
                                // if it is from the same station simply halt the loop, the code
                                // at the end of the loop will replace the old frame with the new
                                check_padding = dev->skb_fake_bad[rp].padding;
                                check_skb = dev->skb_fake_bad[rp].skb;
                                check_wlhdr = (struct ieee80211_hdr*) (check_skb->data + check_padding +
                                        sizeof(struct b43_plcp_hdr6));
                                if(!memcmp(this_wlhdr->addr2, check_wlhdr->addr2, 6))
                                        break;

                                // determine the oldest frame inside the array
                                if(dev->skb_fake_bad[rp].time < mintime) {
                                        minpos = rp;
                                        mintime = dev->skb_fake_bad[rp].time;
                                }
                        }

                        // we only halt the loop if we find a frame from the same station
                        // so the block below is executed only if no frame was found
                        // if a null skb was found use it, otherwise use the oldest
                        if(rp == MAX_SPEC_BLOCKS) {
                                if(nullpos != MAX_SPEC_BLOCKS)
                                        rp = nullpos;
                                else
                                        rp = minpos;
                        }
                } else {
                        rp = 0;
                }

#ifdef MULTI11_DEBUG
                b43dbg(dev->wl, "keeping BAD FRAME, a NACK was sent (packet size = %d)\n",skb->len);
#endif
                dev->nack_sent++;

                if(dev->skb_fake_bad[rp].skb != NULL) {
#ifdef MULTI11_DEBUG
                        b43dbg(dev->wl, "removing old BAD FRAME and keeping new\n");
#endif
                        dev_kfree_skb_any(dev->skb_fake_bad[rp].skb);
                }
                dev->skb_fake_bad[rp].skb = skb;
                dev->skb_fake_bad[rp].time = jiffies;
                dev->skb_fake_bad[rp].padding = padding;
                return;
        }

	plcp = (struct b43_plcp_hdr6 *)(skb->data + padding);
	skb_pull(skb, sizeof(struct b43_plcp_hdr6) + padding);
	/* The skb contains the Wireless Header + payload data now */
	if (unlikely(skb->len < (2 + 2 + 6 /*minimum hdr */  + FCS_LEN))) {
		b43dbg(dev->wl, "RX: Packet size underrun (2)\n");
		goto drop;
	}
	wlhdr = (struct ieee80211_hdr *)(skb->data);

	// GRINGO
	// let's check if this is a special retransmission
	// in this case we should check
	// - we have an old frame
	// - the old frame is the same transmission
	llc = (u8*) (skb->data + 24);
	if(llc[0] == 0x33 && !(macstat & B43_RX_MAC_FCSERR) &&
		!dev->wl->radiotap_enabled &&
		ieee80211_is_data(wlhdr->frame_control)) {
		struct ieee80211_hdr* old_wlhdr;
		u32 blocks, oblocks;
		u32 orig_len, new_len;
		u32 crc32, new_crc32;
		u32 cum_len, cum_len_new;
		u32 blocks_done = 0;
		char* new_pos;
		char* old_pos;
		struct b43_plcp_hdr6* old_plcp;
		u8 old_rate0, old_rate1;

		// leave this value for initialization so that no_action will not
		// freeze!
		int rp = MAX_SPEC_BLOCKS;

		// refresh statistics
		dev->maranello_spec_from_hardware++;

		u8 std_llc[] = {0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00, 0x08, 0x00};

#ifdef MULTI11_DEBUG
		b43dbg(dev->wl, "Special retransmission detected, FCS good\n");
#endif

		// check the new fragment is at least 58+4 bytes long
		// (58 => base block)
		// (4  => FCS)
		new_len = skb->len;
		if(unlikely(new_len < 58 + FCS_LEN)) {
			b43dbg(dev->wl, "Special retransmission is too short\n");
			goto no_action;
		}

		// if we are an AP we must check in the skb pool if we have one with the same source
		// address
		// TODO1 for AP, if we do not find any skb, try reconstruction using the latest skb
		//	check if performances are improved
		if(b43_is_mode(dev->wl, NL80211_IFTYPE_AP)) {
			for(rp = 0; rp < MAX_SPEC_BLOCKS; rp++) {
				struct ieee80211_hdr* check_wlhdr;
				int check_padding;
				struct sk_buff* check_skb;
				if(dev->skb_fake_bad[rp].skb == NULL)
					continue;
				check_padding = dev->skb_fake_bad[rp].padding;
				check_skb = dev->skb_fake_bad[rp].skb;

				// check if we have the same address
				check_wlhdr = (struct ieee80211_hdr*) (check_skb->data + check_padding +
					sizeof(struct b43_plcp_hdr6));

				// check if we have the same addresses
				if(!memcmp(wlhdr->addr2, check_wlhdr->addr2, 6))
					break;
			}

			// if rp is at its maximum we have not found anything, drop the frame
			if(rp == MAX_SPEC_BLOCKS) {
#ifdef MULTI11_DEBUG
				b43dbg(dev->wl, "No original fragment available for this source address\n");
#endif
				goto no_action;
			}
		} else {
			// continue with the first (the only one) skb stored as we are a station
			// check we have a stored skb
			// we can safely jump to no_action (if it is the case) as the skb is null
			rp = 0;
			if(dev->skb_fake_bad[rp].skb == NULL)
				goto no_action;
		}

		// check rp is inside its limits
		if(rp < 0 || rp >= MAX_SPEC_BLOCKS) {
			b43dbg(dev->wl, "rp outside limits\n");
			rp = MAX_SPEC_BLOCKS;
			B43_WARN_ON(1);
			goto no_action;
		}

		// check if the old frame is for the same transmission
		// we should have same addresses and same sequence ctr
		// remember we should remove the initial plcp from the original fragment
		old_plcp = (struct b43_plcp_hdr6 *)(dev->skb_fake_bad[rp].skb->data +
			dev->skb_fake_bad[rp].padding);
		old_rate0 = old_plcp->raw[0];
		old_rate1 = old_plcp->raw[1];
		skb_pull(dev->skb_fake_bad[rp].skb, sizeof(struct b43_plcp_hdr6) + 
			dev->skb_fake_bad[rp].padding);
		
		orig_len = dev->skb_fake_bad[rp].skb->len;
		if(new_len > orig_len) {
			b43dbg(dev->wl, "Special retransmission longer than original fragment\n");
			goto no_action;
		}

		old_wlhdr = (struct ieee80211_hdr*) dev->skb_fake_bad[rp].skb->data;

		// determine which blocks we should replace in the original skb
		blocks = (llc[1] << 16) | (llc[3] << 8) | llc[2];
		oblocks = blocks;

#ifdef MULTI11_DEBUG
		b43dbg(dev->wl, "integrated block = %02X %02X %02X\n", llc[1], llc[3], llc[2]);
#endif

		// replace the base block, keep the crc inside the new llc and replace the
		// llc with a standard one
		crc32 = *(u32*) (skb->data + 28);
		crc32 = skb->data[28] & 0xff;
		crc32 |= (((skb->data[29]) << 8) & 0xff00);
		crc32 |= (((skb->data[30]) << 16) & 0xff0000);
		crc32 |= (((skb->data[31]) << 24) & 0xff000000);
		memcpy(dev->skb_fake_bad[rp].skb->data, skb->data, 58);
		memcpy(dev->skb_fake_bad[rp].skb->data + 24, std_llc, sizeof(std_llc));

		// replace all the retransmitted blocks, check that we are always within the
		// length of the original frame
		old_pos = dev->skb_fake_bad[rp].skb->data + 58;
		new_pos = skb->data + 58;
		cum_len = 58;
		cum_len_new = 58;
		
		// do not copy FCS
		new_len -= 4;
		orig_len -= 4;
#ifdef MULTI11_DEBUG
		b43dbg(dev->wl, "-->len = %d %d\n", new_len, orig_len);
#endif
		
		for(ii=0; ii < 24; ii++)
			tmp_mask_vector[ii] = 0;

		for(ii=0; ii < 24; ii++) {
			int bytes_copy = FLETCHER_BLOCK_SIZE;

			if( blocks & 0x1 ) {
				int bytes_left = new_len - cum_len_new;
				
				if( bytes_left <= 0) {
					b43dbg(dev->wl, "Invalid special retransmission: incorrect mask\n");
					goto no_action;
				}

				if( cum_len + bytes_left > orig_len ) {
					b43dbg(dev->wl, "Special retransmission contains a block outside the original frame\n");
					goto no_action;
				}

				if( bytes_left < FLETCHER_BLOCK_SIZE )
					bytes_copy = bytes_left;

				memcpy(old_pos, new_pos, bytes_copy);

				new_pos += bytes_copy;
				cum_len_new += bytes_copy;

				blocks_done++;
				tmp_mask_vector[ii] = 1;
			}

			old_pos += bytes_copy;
			cum_len += bytes_copy;

			blocks = blocks >> 1;
		}

		//Update statistics
		sum_sigma = 0;
		m = 0;	
		sigma_tmp = 0;
		ones = 0;
	
		for(ii=0; ii < 25; ii++)
		{
			if(tmp_mask_vector[ii] == 1) {	
				dev->special_retr_nth_block[ii]++;
				ones++;
			}
			m += ii*tmp_mask_vector[ii];	
		}

		if(ones > 0) {
			m = m/ones;

			for(ii=0; ii < 25; ii++) {
				sigma_tmp = (ii*tmp_mask_vector[ii])-m;
				sum_sigma += (sigma_tmp*sigma_tmp);
			}
		}

		goto action;

	no_action:
#ifdef MULTI11_DEBUG
		b43dbg(dev->wl, "discarding all...\n");
#endif
		dev_kfree_skb_any(skb);
		if(rp < MAX_SPEC_BLOCKS) {
			if(dev->skb_fake_bad[rp].skb)
				dev_kfree_skb_any(dev->skb_fake_bad[rp].skb);
			//dev->skb_fake_bad[0].skb = NULL;
			dev->skb_fake_bad[rp].skb = NULL;
		}
		dev->maranello_rx_failed++;
		return;

	action:
		// check fcs is correct
#ifdef FLETCHER32
		new_crc32 = fletcher32(dev->skb_fake_bad[rp].skb->data, orig_len, NULL);
#else
		new_crc32 = crc32_be( ~0, dev->skb_fake_bad[rp].skb->data + 32, orig_len - 32);
#endif

		if(new_crc32 != crc32) {
#ifdef MULTI11_DEBUG
			b43dbg(dev->wl, "fcs wrong, discarding all...\n");
#endif
			dev->maranello_rx_fcs_failed++;
			goto no_action;
		}

		// copy the new plcp inside the old one
		memcpy(old_plcp, plcp, sizeof(struct b43_plcp_hdr6));
		plcp = old_plcp;

		// release the new skb
#ifdef MULTI11_DEBUG
		b43dbg(dev->wl, "replacing new skb with old rebuilt\n");
#endif
		dev_kfree_skb_any(skb);
		skb = dev->skb_fake_bad[rp].skb;
		dev->skb_fake_bad[rp].skb = NULL;
		wlhdr = old_wlhdr;

		// refresh statistics
		dev->spread_factor += sum_sigma;
		dev->special_retr++;
		if(blocks_done >=0 && blocks_done <= 24) {
			dev->special_retr_blocks[blocks_done]++;
		}
		else {
			b43dbg(dev->wl, "blocks_done out of limits: %d\n",blocks_done);
		}

	}
	
	/*
	// note: we disabled this code because it breaks hostapd checking!
	// DEBUG, replace addr2 with that of the AP
	// if this is a udp frame generated by our send program
	// then modify the source addresses so that we can see
	// this frame coming out from wlan1 in sta mode
	if(!dev->wl->radiotap_enabled) {
		char APADDR[] = {0x00, 0x14, 0x51, 0x6a, 0x8f, 0x55};

		u8* pkt = (u8*) skb->data;

		if(pkt[32] == 0x45 && // ip
			pkt[32 + 9] == 17 && // udp
			pkt[32 + 28] == 0xde && // first byte of our special sign
			pkt[32 + 29] == 0xad ) // second byte of our special sign
		{
			//memcpy(wlhdr->addr2, APADDR, 6);
			//wlhdr->addr3[0] = get_speed_from_encoding(plcp->raw[0], plcp->raw[1]) | 0x80;
		}
	}
	*/

	fctl = wlhdr->frame_control;

	if (macstat & B43_RX_MAC_DEC) {
		unsigned int keyidx;
		int wlhdr_len;

		keyidx = ((macstat & B43_RX_MAC_KEYIDX)
			  >> B43_RX_MAC_KEYIDX_SHIFT);
		/* We must adjust the key index here. We want the "physical"
		 * key index, but the ucode passed it slightly different.
		 */
		keyidx = b43_kidx_to_raw(dev, keyidx);
		B43_WARN_ON(keyidx >= dev->max_nr_keys);

		if (dev->key[keyidx].algorithm != B43_SEC_ALGO_NONE) {
			wlhdr_len = ieee80211_hdrlen(fctl);
			if (unlikely(skb->len < (wlhdr_len + 3))) {
				b43dbg(dev->wl,
				       "RX: Packet size underrun (3)\n");
				goto drop;
			}
			status.flag |= RX_FLAG_DECRYPTED;
		}
	}

	/* Link quality statistics */
	status.noise = dev->stats.link_noise;
	if ((chanstat & B43_RX_CHAN_PHYTYPE) == B43_PHYTYPE_N) {
//		s8 rssi = max(rxhdr->power0, rxhdr->power1);
		//TODO: Find out what the rssi value is (dBm or percentage?)
		//      and also find out what the maximum possible value is.
		//      Fill status.ssi and status.signal fields.
	} else {
		status.signal = b43_rssi_postprocess(dev, rxhdr->jssi,
						  (phystat0 & B43_RX_PHYST0_OFDM),
						  (phystat0 & B43_RX_PHYST0_GAINCTL),
						  (phystat3 & B43_RX_PHYST3_TRSTATE));
		status.qual = (rxhdr->jssi * 100) / B43_RX_MAX_SSI;
		//printk("Signal %d, quality %d\n",status.signal,status.qual);
	}

	if (phystat0 & B43_RX_PHYST0_OFDM)
		status.rate_idx = b43_plcp_get_bitrate_idx_ofdm(plcp,
						phytype == B43_PHYTYPE_A);
	else
		status.rate_idx = b43_plcp_get_bitrate_idx_cck(plcp);
	if (unlikely(status.rate_idx == -1)) {
		/* PLCP seems to be corrupted.
		 * Drop the frame, if we are not interested in corrupted frames. */
		if (!(dev->wl->filter_flags & FIF_PLCPFAIL))
			goto drop;
	}
	status.antenna = !!(phystat0 & B43_RX_PHYST0_ANT);

	/*
	 * All frames on monitor interfaces and beacons always need a full
	 * 64-bit timestamp. Monitor interfaces need it for diagnostic
	 * purposes and beacons for IBSS merging.
	 * This code assumes we get to process the packet within 16 bits
	 * of timestamp, i.e. about 65 milliseconds after the PHY received
	 * the first symbol.
	 */
	if (ieee80211_is_beacon(fctl) || dev->wl->radiotap_enabled) {
		u16 low_mactime_now;

		b43_tsf_read(dev, &status.mactime);
		low_mactime_now = status.mactime;
		status.mactime = status.mactime & ~0xFFFFULL;
		status.mactime += mactime;
		if (low_mactime_now <= mactime)
			status.mactime -= 0x10000;
		status.flag |= RX_FLAG_TSFT;
	}

	chanid = (chanstat & B43_RX_CHAN_ID) >> B43_RX_CHAN_ID_SHIFT;
	switch (chanstat & B43_RX_CHAN_PHYTYPE) {
	case B43_PHYTYPE_A:
		status.band = IEEE80211_BAND_5GHZ;
		B43_WARN_ON(1);
		/* FIXME: We don't really know which value the "chanid" contains.
		 *        So the following assignment might be wrong. */
		status.freq = b43_channel_to_freq_5ghz(chanid);
		break;
	case B43_PHYTYPE_G:
		status.band = IEEE80211_BAND_2GHZ;
		/* chanid is the radio channel cookie value as used
		 * to tune the radio. */
		status.freq = chanid + 2400;
		break;
	case B43_PHYTYPE_N:
		/* chanid is the SHM channel cookie. Which is the plain
		 * channel number in b43. */
		if (chanstat & B43_RX_CHAN_5GHZ) {
			status.band = IEEE80211_BAND_5GHZ;
			status.freq = b43_freq_to_channel_5ghz(chanid);
		} else {
			status.band = IEEE80211_BAND_2GHZ;
			status.freq = b43_freq_to_channel_2ghz(chanid);
		}
		break;
	default:
		B43_WARN_ON(1);
		goto drop;
	}

	memcpy(IEEE80211_SKB_RXCB(skb), &status, sizeof(status));
	ieee80211_rx_irqsafe(dev->wl->hw, skb);

	return;
drop:
	b43dbg(dev->wl, "RX: Packet dropped\n");
	dev_kfree_skb_any(skb);
}

void b43_handle_txstatus(struct b43_wldev *dev,
			 const struct b43_txstatus *status)
{
	b43_debugfs_log_txstat(dev, status);

	if (status->intermediate)
		return;
	if (status->for_ampdu)
		return;
	if (!status->acked)
		dev->wl->ieee_stats.dot11ACKFailureCount++;
	if (status->rts_count) {
		if (status->rts_count == 0xF)	//FIXME
			dev->wl->ieee_stats.dot11RTSFailureCount++;
		else
			dev->wl->ieee_stats.dot11RTSSuccessCount++;
	}

	if (b43_using_pio_transfers(dev))
		b43_pio_handle_txstatus(dev, status);
	else
		b43_dma_handle_txstatus(dev, status);

	b43_phy_txpower_check(dev, 0);
}

/* Fill out the mac80211 TXstatus report based on the b43-specific
 * txstatus report data. This returns a boolean whether the frame was
 * successfully transmitted. */
bool b43_fill_txstatus_report(struct b43_wldev *dev,
			      struct ieee80211_tx_info *report,
			      const struct b43_txstatus *status)
{
	bool frame_success = 1;
	int retry_limit;

	/* preserve the confiured retry limit before clearing the status
	 * The xmit function has overwritten the rc's value with the actual
	 * retry limit done by the hardware */
	retry_limit = report->status.rates[0].count;
	ieee80211_tx_info_clear_status(report);

	if (status->acked) {
		/* The frame was ACKed. */
		report->flags |= IEEE80211_TX_STAT_ACK;
	} else {
		/* The frame was not ACKed... */
		if (!(report->flags & IEEE80211_TX_CTL_NO_ACK)) {
			/* ...but we expected an ACK. */
			frame_success = 0;
		}
	}
	if (status->frame_count == 0) {
		/* The frame was not transmitted at all. */
		report->status.rates[0].count = 0;
	} else if (status->rts_count > dev->wl->hw->conf.short_frame_max_tx_count) {
		/*
		 * If the short retries (RTS, not data frame) have exceeded
		 * the limit, the hw will not have tried the selected rate,
		 * but will have used the fallback rate instead.
		 * Don't let the rate control count attempts for the selected
		 * rate in this case, otherwise the statistics will be off.
		 */
		report->status.rates[0].count = 0;
		report->status.rates[1].count = status->frame_count;
	} else {
		if (status->frame_count > retry_limit) {
			report->status.rates[0].count = retry_limit;
			report->status.rates[1].count = status->frame_count -
					retry_limit;

		} else {
			report->status.rates[0].count = status->frame_count;
			report->status.rates[1].idx = -1;
		}
	}

	return frame_success;
}

/* Stop any TX operation on the device (suspend the hardware queues) */
void b43_tx_suspend(struct b43_wldev *dev)
{
	if (b43_using_pio_transfers(dev))
		b43_pio_tx_suspend(dev);
	else
		b43_dma_tx_suspend(dev);
}

/* Resume any TX operation on the device (resume the hardware queues) */
void b43_tx_resume(struct b43_wldev *dev)
{
	if (b43_using_pio_transfers(dev))
		b43_pio_tx_resume(dev);
	else
		b43_dma_tx_resume(dev);
}
